﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Runtime.InteropServices;
using System.Threading;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Forms;

namespace VInjector
{
    // ----------------------------------------------------------------------------------------------------------
    // This tool receives Keyboard/Mouse/Touch events sent by Ventuz that have been collected by 
    // the 'Interaction Rectangle' node and injects it to the Windows operation system.
    // So you need two systems: sender and receiver. It won't work if sender and receiver are on the same system.
    // ----------------------------------------------------------------------------------------------------------

    public partial class MainWindow : Window
    {
        #region members

        VIInjector vInject = null;

        #endregion

        #region Constructor
        public MainWindow()
        {
            InitializeComponent();

            // In case of trouble enable:
            //DebugLogWriter.Init(@"D:\DEV\VInjector\touchInject.log", false, 0);         

            this.vInject = new VIInjector(false);
            this.DataContext = this.vInject;
        }

        protected override void OnClosing(System.ComponentModel.CancelEventArgs e)
        {
            this.vInject.StopReceiver();

            base.OnClosing(e);
        }
        #endregion

        #region UI Events
        private void ReceiveButton_Click(object sender, RoutedEventArgs e)
        {
            if ( sender == this.bt_Receive )
            {
                if ( this.bt_Receive.IsChecked == true )
                    this.vInject.StartReceiver();
                else
                    this.vInject.StopReceiver();
            }
        }

        private void Cb_Device_Checked(object sender, RoutedEventArgs e)
        {
            var dev = sender as System.Windows.Controls.CheckBox;

            if ( dev != null )
            {
                if ( (string)dev.Content == "Mouse" )
                    this.vInject.mouse = dev.IsChecked == true;
                else if ( (string)dev.Content == "Keyboard" )
                    this.vInject.keys = dev.IsChecked == true;
                else if ( (string)dev.Content == "Touch" )
                    this.vInject.touch = dev.IsChecked == true;
            }
        }

        private void Cb_ID_Checked(object sender, RoutedEventArgs e)
        {
            var dev = sender as System.Windows.Controls.CheckBox;

            if ( dev != null )
            {
                if ( (string)dev.Content == "0" )
                {
                    if ( dev.IsChecked == true )
                        this.vInject.activeIDs.Add(0);
                    else
                        this.vInject.activeIDs.Remove(0);
                }
                else if ( (string)dev.Content == "1" )
                {
                    if ( dev.IsChecked == true )
                        this.vInject.activeIDs.Add(1);
                    else
                        this.vInject.activeIDs.Remove(1);
                }
                else if ( (string)dev.Content == "2" )
                {
                    if ( dev.IsChecked == true )
                        this.vInject.activeIDs.Add(2);
                    else
                        this.vInject.activeIDs.Remove(2);
                }
                else if ( (string)dev.Content == "3" )
                {
                    if ( dev.IsChecked == true )
                        this.vInject.activeIDs.Add(3);
                    else
                        this.vInject.activeIDs.Remove(3);
                }
                else if ( (string)dev.Content == "4" )
                {
                    if ( dev.IsChecked == true )
                        this.vInject.activeIDs.Add(4);
                    else
                        this.vInject.activeIDs.Remove(4);
                }
            }
        }

        #endregion
    }


    #region class VIInjector

    /// <summary>
    /// Ventuz Interaction Injection
    /// </summary>
    public class VIInjector
    {
        #region Constants

        const string ID_TouchAdd =          "TouchAdd";
        const string ID_TouchUpdate =       "TouchMove";
        const string ID_TouchRemove =       "TouchRemove";
        const string ID_MouseDoubleClick =  "DoubleClick";          // empty atom
        const string ID_MouseDoubleClick2 = "DoubleClick2";         // int atom containing the button that has double clicked
        const string ID_MouseHover =        "Hover";                // empty atom
        const string ID_MouseIsInside =     "IsInside";             // seems not to be set
        const string ID_MouseButtonDown =   "ButtonDown";
        const string ID_MouseButtonUp =     "ButtonUp";
        const string ID_MousePosition =     "Position";
        const string ID_MouseWheel =        "Wheel";
        const string ID_KeyDown =           "KeyDown";
        const string ID_KeyPress =          "KeyPress";
        const string ID_KeyUp =             "KeyUp";

        const string IP = "225.225.225.1";
        const string PORT = "19344";

        #endregion

        readonly static int Win32InputSize = Marshal.SizeOf(typeof(Win32.INPUT));


        internal bool mouse = true;
        internal bool keys = true;
        internal bool touch = true;

        List<Screen> screenSelector = new List<Screen>();
        internal SortedSet<int> activeIDs = new SortedSet<int>(new int[] { 0 });
        Rectangle targetArea;
        uint maxTouchPoints = 128;
        uint frameCount;

        public string Address => $"{IP}:{PORT}";

        public VIInjector(bool showFeedback)
        {
            if ( !InitTouch(showFeedback) )
                DebugLogWriter.Enqueue("{0}", "Touch Injection Initialization Failed!");

            ScanScreens();
        }

        private bool InitTouch(bool showFeedback)
        {
            return Win32.NativeMethods.InitializeTouchInjection(maxTouchPoints, (int)(showFeedback ? Win32.TOUCHFEEDBACKTYPE.TOUCH_FEEDBACK_INDIRECT : Win32.TOUCHFEEDBACKTYPE.TOUCH_FEEDBACK_NONE));
        }

        private void ScanScreens()
        {
            foreach ( MonitorInfo mi in MonitorInfo.GetAllMonitors() )
                screenSelector.Add(mi.Screen);

            targetArea = screenSelector[0].Bounds;
        }

        #region Touches

        /// <summary>
        /// This method does the final touch injection into the Windows OS.
        /// It transforms the received touch information in the required structures and calls the injection method.
        /// For details see:
        /// https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-initializetouchinjection
        /// https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-injecttouchinput
        /// </summary>
        /// <param name="touches">Queue of touches per ID</param>
        internal void InjectTouches(SortedDictionary<int, Queue<VIInjector.TouchInfo>> touches)
        {
            if ( touches == null || touches.Count == 0 )
                return;

            bool finished = false;
            int radius = 2;

            while ( !finished )
            {
                List<Win32.POINTERTOUCHINFO> injectList = new List<Win32.POINTERTOUCHINFO>();
                List<int> removeList = new List<int>();
                bool lastEntry = true;

                try
                {
                    foreach ( var kvPair in touches )
                    {
                        VIInjector.TouchInfo touchInfo = null;

                        // we need to keep the last entry for updates
                        if ( kvPair.Value.Count > 1 )
                        {
                            touchInfo = kvPair.Value.Dequeue();
                            lastEntry = false;
                        }
                        else if ( kvPair.Value.Count == 1 )
                        {
                            touchInfo = kvPair.Value.Peek();
                            lastEntry &= true;
                        }

                        // map relative to absolute screen coordinates
                        int[] injectCoords = ToInjectionCoordinates((float)touchInfo.PosX, (float)touchInfo.PosY);

                        // make a new pointertouchinfo with all neccessary information
                        Win32.POINTERTOUCHINFO pti = new Win32.POINTERTOUCHINFO();
                        pti.pointerInfo.pointerType = Win32.POINTERINPUTTYPE.PT_TOUCH;
                        pti.touchFlags = Win32.TOUCHFLAGS.TOUCH_FLAGS_NONE;
                        //pti.PointerInfo.Orientation = (uint)cur.getAngleDegrees();    // this is only valid for TuioObjects

                        pti.pressure = 1024;
                        pti.touchMasks = (uint)(Win32.TOUCHMASK.TOUCH_MASK_CONTACTAREA | Win32.TOUCHMASK.TOUCH_MASK_ORIENTATION | Win32.TOUCHMASK.TOUCH_MASK_PRESSURE);
                        pti.pointerInfo.ptPixelLocation.x = injectCoords[0];
                        pti.pointerInfo.ptPixelLocation.y = injectCoords[1];
                        pti.pointerInfo.pointerId = touchInfo.PointerID;
                        //pti.PointerInfo.FrameId = frameCount;

                        pti.rcContact.left = injectCoords[0] - radius;
                        pti.rcContact.right = injectCoords[0] + radius;
                        pti.rcContact.top = injectCoords[1] - radius;
                        pti.rcContact.bottom = injectCoords[1] + radius;

                        // set the right flags
                        if ( touchInfo.TouchRemove == false )
                        {
                            // ADD
                            touchInfo.TouchRemove = null;   // if we do not get an update in next 'frame', update with current possition; additinally we prevent this way a multiple Touch Down injection

                            pti.pointerInfo.pointerFlags = Win32.POINTERFLAGS.POINTER_FLAG_DOWN | Win32.POINTERFLAGS.POINTER_FLAG_INRANGE | Win32.POINTERFLAGS.POINTER_FLAG_INCONTACT;

                            DebugLogWriter.Enqueue("Touch {0}: @({1} x {2}) ID = {3} ({4}) [{5}]", "DOWN", injectCoords[0], injectCoords[1], touchInfo.TouchID, pti.pointerInfo.pointerId, frameCount);
                        }
                        else if ( touchInfo.TouchRemove == null )
                        {
                            // UPDATE
                            pti.pointerInfo.pointerFlags = Win32.POINTERFLAGS.POINTER_FLAG_UPDATE | Win32.POINTERFLAGS.POINTER_FLAG_INRANGE | Win32.POINTERFLAGS.POINTER_FLAG_INCONTACT;

                            DebugLogWriter.Enqueue("Touch {0}: @({1} x {2}) ID = {3} ({4}) [{5}]", "Update", injectCoords[0], injectCoords[1], touchInfo.TouchID, pti.pointerInfo.pointerId, frameCount);
                        }
                        else if ( touchInfo.TouchRemove == true )
                        {
                            // REMOVE
                            pti.pointerInfo.pointerFlags = Win32.POINTERFLAGS.POINTER_FLAG_UP;
                            removeList.Add(kvPair.Key);

                            DebugLogWriter.Enqueue("Touch {0}: @({1} x {2}) ID = {3} ({4}) [{5}]", "UP", injectCoords[0], injectCoords[1], touchInfo.TouchID, pti.pointerInfo.pointerId, frameCount);
                        }

                        // add it to injection list
                        injectList.Add(pti);

                    }   // foreach

                    finished = lastEntry;
                }
                catch ( Exception ex )
                {
                    DebugLogWriter.Enqueue("generating injection list failed: {0}", ex.Message);
                }

                // inject touch events
                if ( injectList.Count > 0 )
                {
                    if ( !Win32.NativeMethods.InjectTouchInput((uint)injectList.Count, injectList.ToArray()) )
                    {
                        uint err = Win32.NativeMethods.GetLastError();

                        DebugLogWriter.Enqueue("Touch Inject Error: {0}", err);
                    }
                }

                if ( removeList.Count > 0 )
                {
                    // remove TouchInfo queue for those touches that received an Up event
                    foreach ( var key in removeList )
                    {
                        var v = touches[key];
                        touches.Remove(key);

                        DebugLogWriter.Enqueue("Removing Touch with ID = {0} ({1})", v.Peek().TouchID, key);
                    }
                }

                // count up
                frameCount++;
            }   // while
        }

        private int[] ToInjectionCoordinates(float x, float y)
        {
            int[] result = new int[2];
            result[0] = targetArea.X + (int)Math.Round(x * targetArea.Width);
            result[1] = targetArea.Y + (int)Math.Round(y * targetArea.Height);
            return result;
        }

        #endregion

        #region Mouse

        public static void InjectMouseButton(VIInjector.MouseButtons mouseButton, bool up)
        {
            Win32.INPUT[] mouseInputs = new Win32.INPUT[2];

            Win32.INPUT mi = new Win32.INPUT();
            mi.Type = (uint)Win32.SENDINPUTEVENTTYPE.InputMouse;
            mi.Data = new Win32.MOUSEKEYBDHARDWAREINPUT();
            mi.Data.Mouse = new Win32.MOUSEINPUT();

            switch ( mouseButton )
            {
                case VIInjector.MouseButtons.None:
                    break;
                case VIInjector.MouseButtons.Left:
                    if ( !up ) mi.Data.Mouse.Flags = (uint)Win32.MOUSEFLAGS.MOUSEEVENT_LEFTDOWN;
                    else mi.Data.Mouse.Flags = (uint)Win32.MOUSEFLAGS.MOUSEEVENT_LEFTUP;
                    break;

                case VIInjector.MouseButtons.Right:
                    if ( !up ) mi.Data.Mouse.Flags = (uint)Win32.MOUSEFLAGS.MOUSEEVENT_RIGHTDOWN;
                    else mi.Data.Mouse.Flags = (uint)Win32.MOUSEFLAGS.MOUSEEVENT_RIGHTUP;
                    break;

                case VIInjector.MouseButtons.Middle:
                    if ( !up ) mi.Data.Mouse.Flags = (uint)Win32.MOUSEFLAGS.MOUSEEVENT_MIDDLEDOWN;
                    else mi.Data.Mouse.Flags = (uint)Win32.MOUSEFLAGS.MOUSEEVENT_MIDDLEUP;
                    break;

                case VIInjector.MouseButtons.XButton1:
                    if ( !up ) mi.Data.Mouse.Flags = (uint)Win32.MOUSEFLAGS.MOUSEEVENT_XDOWN;
                    else mi.Data.Mouse.Flags = (uint)Win32.MOUSEFLAGS.MOUSEEVENT_XUP;
                    break;
                case VIInjector.MouseButtons.XButton2:
                    // TODO: missing
                    break;
                default:
                    break;
            }

            mouseInputs[0] = mi;

            uint result = Win32.NativeMethods.SendInput(1, mouseInputs, Win32InputSize);

            if ( result == 0 )
                DebugLogWriter.Enqueue("Mouse Button injection failed!");
        }

        public static void InjectMouseWheel(int delta)
        {
            Win32.INPUT[] mouseInputs = new Win32.INPUT[2];
            Win32.INPUT mi = new Win32.INPUT();
            mi.Type = (uint)Win32.SENDINPUTEVENTTYPE.InputMouse;
            mi.Data = new Win32.MOUSEKEYBDHARDWAREINPUT();
            mi.Data.Mouse = new Win32.MOUSEINPUT();

            mi.Data.Mouse.Flags = (uint)Win32.MOUSEFLAGS.MOUSEEVENT_WHEEL;
            mi.Data.Mouse.MouseData = (uint)delta;

            mouseInputs[0] = mi;

            uint result = Win32.NativeMethods.SendInput(1, mouseInputs, Win32InputSize);

            if ( result == 0 )
                DebugLogWriter.Enqueue("Mouse Wheel injection failed!");

        }

        public static void InjectMouseMove(double x, double y)
        {
            Win32.INPUT[] mouseInputs = new Win32.INPUT[2];
            Win32.INPUT mi = new Win32.INPUT();
            mi.Type = (uint)Win32.SENDINPUTEVENTTYPE.InputMouse;
            mi.Data = new Win32.MOUSEKEYBDHARDWAREINPUT();
            mi.Data.Mouse = new Win32.MOUSEINPUT();

            mi.Data.Mouse.X = (int)(x * 65536.0);
            mi.Data.Mouse.Y = (int)(y * 65536.0);
            mi.Data.Mouse.Flags = (uint)(Win32.MOUSEFLAGS.MOUSEEVENT_MOVE | Win32.MOUSEFLAGS.MOUSEEVENT_ABSOLUTE);

            mouseInputs[0] = mi;

            uint result = Win32.NativeMethods.SendInput(1, mouseInputs, Win32InputSize);

            if ( result == 0 )
                DebugLogWriter.Enqueue("Mouse Move injection failed!");

        }

        #endregion

        #region Keyboard

        public static void InjectKeyPress(int keyCode)
        {
            List<Win32.INPUT> inputs = new List<Win32.INPUT>();

            Win32.INPUT down = new Win32.INPUT();
            down.Type = (uint)Win32.SENDINPUTEVENTTYPE.InputKeyboard;
            down.Data.Keyboard = new Win32.KEYBDINPUT();
            down.Data.Keyboard.KeyCode = (UInt16)keyCode;
            down.Data.Keyboard.Time = 0;
            down.Data.Keyboard.ExtraInfo = IntPtr.Zero;

            // Handle extended keys:
            // If the scan code is preceded by a prefix byte that has the value 0xE0 (224),
            // we need to include the KEYEVENTF_EXTENDEDKEY flag in the Flags property. 
            if ( (keyCode & 0xFF00) == 0xE000 )
                down.Data.Keyboard.Flags |= (UInt32)Win32.KEYBOARDFLAGS.ExtendedKey;

            inputs.Add(down);
            Win32.NativeMethods.SendInput((UInt32)inputs.Count, inputs.ToArray(), Win32InputSize);
        }

        public static void InjectKeyRelease(int keyCode)
        {
            List<Win32.INPUT> inputs = new List<Win32.INPUT>();

            Win32.INPUT up = new Win32.INPUT();
            up.Type = (uint)Win32.SENDINPUTEVENTTYPE.InputKeyboard;
            up.Data.Keyboard = new Win32.KEYBDINPUT();
            up.Data.Keyboard.KeyCode = (UInt16)keyCode;
            up.Data.Keyboard.Flags = (UInt32)(Win32.KEYBOARDFLAGS.KeyUp);
            up.Data.Keyboard.Time = 0;
            up.Data.Keyboard.ExtraInfo = IntPtr.Zero;

            // Handle extended keys:
            // If the scan code is preceded by a prefix byte that has the value 0xE0 (224),
            // we need to include the KEYEVENTF_EXTENDEDKEY flag in the Flags property. 
            if ( (keyCode & 0xFF00) == 0xE000 )
                up.Data.Keyboard.Flags |= (UInt32)Win32.KEYBOARDFLAGS.ExtendedKey;

            inputs.Add(up);
            Win32.NativeMethods.SendInput((UInt32)inputs.Count, inputs.ToArray(), Win32InputSize);
        }

        #endregion

        #region Receiving from Ventuz

        #region enum DeviceTypes

        [Flags]
        public enum DeviceTypes
        {
            Unknown = 0x0000,
            MultiTouch = 0x0001,
            Mouse = 0x0002,
            Keyboard = 0x0004,
            RawMouse = 0x0008,
            RawKeyboard = 0x0010,
            GamePad = 0x0020,
            VRController = 0x0040,

            Filtered = 0x0100,

            TouchFromWindowsMouse = Filtered | MultiTouch,
            WindowsMouseFromTouch = Filtered | Mouse,
        }

        #endregion

        #region enum MouseButtons

        [Flags]
        public enum MouseButtons
        {
            None = 0,
            Left = 1048576,
            Right = 2097152,
            Middle = 4194304,
            XButton1 = 8388608,
            XButton2 = 16777216
        }

        #endregion

        #region internal class TouchInfo

        internal class TouchInfo
        {
            internal int SenderID;
            internal int TouchID;       // from Ventuz
            internal uint PointerID;    // for injection
            internal bool? TouchRemove; // false = Add; null = Update; true = Remove
            internal double PosX;
            internal double PosY;
        } 

        #endregion

        SortedDictionary<int, Queue<TouchInfo>> touchInfos = new SortedDictionary<int, Queue<TouchInfo>>();
        SortedSet<int> activeTouchIDs = new SortedSet<int>();

        Ventuz.OSC.UdpReader udpReader = null;
        CancellationTokenSource cts = new CancellationTokenSource();

        MouseButtons mButton = MouseButtons.None;

        #region Start / Stop

        internal void StartReceiver()
        {
            cts = new CancellationTokenSource();
            Task.Factory.StartNew(() => Receive(cts.Token));
        }

        internal void StopReceiver()
        {
            try
            {
                if ( cts != null )
                    cts.Cancel();
                cts = null;
            }
            catch { }
        }

        #endregion

        private async Task Receive(CancellationToken ct)
        {
            if ( ct == null )
                throw new NullReferenceException("ct");

            if ( udpReader == null )
                udpReader = new Ventuz.OSC.UdpReader(19344, "225.225.225.1");

            while ( true )
            {
                if ( udpReader.Poll(100) )
                {
                    try
                    {
                        #region Message Structure

                        // - OscBundle
                        // -- OscElement (Adr/TC)
                        // -- OscElement[] (EventAtom) 

                        // The message format is documented here: https://www.ventuz.com/support/help/latest/NodeInteractionInteractionRect.html

                        #endregion

                        var oscMsg = udpReader.Receive();

                        if ( oscMsg is Ventuz.OSC.OscBundle bundle )
                        {
                            bool first = true;
                            PrepareTouchLists();

                            foreach ( var item in bundle.Elements )
                            {
                                if ( item is Ventuz.OSC.OscElement elmt )
                                {
                                    if ( first )
                                    {
                                        if ( !elmt.Address.StartsWith("/ventuz") )
                                            break;
                                        first = false;
                                    }
                                    else
                                    {
                                        #region Interaction Data Types

                                        // As second element we expect a bundle with the actual interaction data.
                                        // Possible messages:   Address= /ventuz/<ReceiverID>(int)/<DeviceType>(string)/<DeviceID(int)>/<ID>(string)/<TouchGroup(int)>
                                        // Touch:               Args=    /<TouchType>(int)/<TouchID(string)>/<X>(float)/<Y>(float)/<Z>(float)/<Pressure>(float)/<Angle>(float)
                                        // Mouse:               Args=    /<X>(float)/<Y>(float)/<Z>(float)
                                        //                      Args=    /<Value>(object[1])
                                        //                      Args=    NONE
                                        // Keyboard:            Args=    /<Value>(object[1]) 

                                        #endregion

                                        if ( item is Ventuz.OSC.OscElement elmt2 )
                                        {
                                            string adr = elmt2.Address;
                                            string[] adrTokens = adr.Split(new string[] { "/" }, StringSplitOptions.RemoveEmptyEntries);

                                            if ( adrTokens.Length > 2 && int.TryParse(adrTokens[1], out int msgID) && this.activeIDs.Contains(msgID) && Enum.TryParse(adrTokens[2], out DeviceTypes devType) )
                                            {
                                                object[] args = elmt2.Args;

                                                switch ( devType )
                                                {
                                                    case DeviceTypes.MultiTouch when touch == true:
                                                        {
                                                            #region Touch
                                                            //Console.WriteLine("DevType: MultiTouch");

                                                            int.TryParse((string)args[1], out int touchID);
                                                            float x = (float)args[2];
                                                            float y = (float)args[3];
                                                            string interactionID = adrTokens[4] as string;

                                                            if ( interactionID != null )
                                                            {
                                                                TouchInfo ti;
                                                                int pid;
                                                                Queue<TouchInfo> tiQueue;

                                                                switch ( interactionID )
                                                                {
                                                                    case ID_TouchAdd:

                                                                        pid = GetNextLowestID();

                                                                        ti = new TouchInfo() { SenderID = msgID, TouchID = touchID, PointerID = (uint)pid, PosX = x, PosY = y, TouchRemove = false };
                                                                        tiQueue = new Queue<TouchInfo>();
                                                                        tiQueue.Enqueue(ti);

                                                                        touchInfos.Add(touchID, tiQueue);

                                                                        activeTouchIDs.Add(touchID);

                                                                        break;

                                                                    case ID_TouchUpdate:

                                                                        if ( activeTouchIDs.Contains(touchID) )
                                                                        {
                                                                            tiQueue = touchInfos[touchID];
                                                                            pid = (int)tiQueue.Peek().PointerID;

                                                                            ti = new TouchInfo() { SenderID = msgID, TouchID = touchID, PointerID = (uint)pid, PosX = x, PosY = y, TouchRemove = null };
                                                                            tiQueue.Enqueue(ti);
                                                                        }
                                                                        else
                                                                        {
                                                                            // we did not receive an Add/Down event for this touch ID update - catch up here to prevent injection errors!
                                                                            pid = GetNextLowestID();

                                                                            ti = new TouchInfo() { SenderID = msgID, TouchID = touchID, PointerID = (uint)pid, PosX = x, PosY = y, TouchRemove = false };
                                                                            tiQueue = new Queue<TouchInfo>();
                                                                            tiQueue.Enqueue(ti);

                                                                            touchInfos.Add(touchID, tiQueue);

                                                                            activeTouchIDs.Add(touchID);
                                                                        }

                                                                        break;

                                                                    case ID_TouchRemove:

                                                                        tiQueue = touchInfos[touchID];
                                                                        pid = (int)tiQueue.Peek().PointerID;

                                                                        ti = new TouchInfo() { SenderID = msgID, TouchID = touchID, PointerID = (uint)pid, PosX = x, PosY = y, TouchRemove = true };
                                                                        tiQueue.Enqueue(ti);

                                                                        activeTouchIDs.Remove(touchID);

                                                                        break;

                                                                    default:
                                                                        break;
                                                                }
                                                            }

                                                            break;
                                                            #endregion
                                                        }
                                                    case DeviceTypes.Mouse when mouse == true:
                                                        {
                                                            #region Mouse

                                                            string interactionID = adrTokens[4] as string;

                                                            if ( interactionID != null )
                                                            {
                                                                switch ( interactionID )
                                                                {
                                                                    case ID_MouseButtonDown:

                                                                        this.mButton |= (MouseButtons)args[0];

                                                                        InjectMouseButton((MouseButtons)args[0], false);

                                                                        break;

                                                                    case ID_MousePosition:

                                                                        float x = (float)args[0];
                                                                        float y = (float)args[1];

                                                                        InjectMouseMove(x, y);

                                                                        break;

                                                                    case ID_MouseButtonUp:

                                                                        MouseButtons up = (MouseButtons)args[0];
                                                                        this.mButton &= ~up;

                                                                        InjectMouseButton((MouseButtons)args[0], true);

                                                                        break;

                                                                    case ID_MouseWheel:

                                                                        if ( int.TryParse(args[0].ToString(), out int wheelDelta) )
                                                                        {
                                                                            InjectMouseWheel(wheelDelta);
                                                                        }

                                                                        break;

                                                                    case ID_MouseHover:


                                                                        break;

                                                                    case ID_MouseDoubleClick:


                                                                        break;

                                                                    case ID_MouseDoubleClick2:


                                                                        break;

                                                                    default:
                                                                        break;
                                                                }
                                                            }

                                                            break;
                                                            #endregion
                                                        }
                                                    case DeviceTypes.Keyboard when keys == true:
                                                        {
                                                            #region Keyboard

                                                            string interactionID = adrTokens[4] as string;
                                                            int ikey;

                                                            if ( interactionID != null )
                                                            {
                                                                switch ( interactionID )
                                                                {
                                                                    case ID_KeyDown:

                                                                        ikey = (int)args[0];

                                                                        InjectKeyPress(ikey);

                                                                        break;

                                                                    case ID_KeyUp:

                                                                        ikey = (int)args[0];

                                                                        InjectKeyRelease(ikey);

                                                                        break;

                                                                    case ID_KeyPress:

                                                                        break;

                                                                    default:
                                                                        break;
                                                                }
                                                            }

                                                            break;
                                                            #endregion
                                                        }
                                                    default:
                                                        break;
                                                }
                                            }
                                        }
                                    }
                                }
                            }   // foreach

                            if ( touchInfos.Count > 0 )
                            {
                                this.InjectTouches(touchInfos);
                            }
                        }
                    }
                    catch ( Exception ex )
                    {
                        DebugLogWriter.Enqueue("Input Data Receive Error: {0}", ex.Message);
                        throw ex;
                    }
                }

                if ( ct.IsCancellationRequested )
                    throw new TaskCanceledException();

                await Task.Delay(10);
            }   // while
        }

        /// <summary>
        /// Returns lowest free pointer ID for injection
        /// </summary>
        private int GetNextLowestID()
        {
            if ( this.touchInfos == null || touchInfos.Count == 0 )
                return 0;

            int id = 0;

            foreach ( var kvpair in touchInfos )
            {
                if ( id < kvpair.Value.Peek().PointerID )
                    break;
                else
                    id++;
            }

            return id;
        }

        private void PrepareTouchLists()
        {
            // always keep one 'update' per active touch until it is really removed by touch up
            foreach ( var kvPair in touchInfos )
            {
                var list = kvPair.Value;

                if ( list.Count > 0 )
                {
                    var ti = list.Dequeue();
                    list.Clear();
                    // make sure that remaining TouchInfo is an 'update'
                    ti.TouchRemove = null;
                    list.Enqueue(ti);
                }
                else
                {
                    System.Diagnostics.Debug.Fail("TouchInfo list is empty!");
                }
            }
        }

        #endregion
    }

    #endregion


    #region clapp MonitorInfo

    public class MonitorInfo
    {
        public string FriendlyName { get; set; }
        public Screen Screen { get; set; }

        public static MonitorInfo[] GetAllMonitors()
        {
            List<MonitorInfo> monitors = new List<MonitorInfo>();
            foreach ( Screen s in Screen.AllScreens )
            {
                MonitorInfo mi = new MonitorInfo();
                mi.Screen = s;
                mi.FriendlyName = s.DeviceName.Replace("\\\\.\\DISPLAY", "Monitor ") + string.Format(" ({0}x{1})", s.Bounds.Width, s.Bounds.Height);
                monitors.Add(mi);
            }
            return monitors.ToArray();
        }
    }

    #endregion

    #region class DebugLogWriter

    /// <summary>
    /// Item to written by the DebugLogWriter class.
    /// </summary>
    public struct DebugLogItem
    {
        public string Format;
        public object[] Entries;
        string Date;

        public DebugLogItem(string format, params object[] entries)
        {
            this.Format = format;
            this.Entries = entries;
            this.Date = DateTime.Now.ToLocalTime().ToString("hh:mm:ss.fff");
        }

        public override string ToString()
        {
            if ( this.Format != null && this.Format.Length > 0 && this.Entries != null )
            {
                try
                {
                    string s = $"{Date}: {string.Format(this.Format, this.Entries)}";
                    return s;
                }
                catch
                {
                    return "<BAD FORMAT>";
                }
            }

            return base.ToString();
        }
    }

    /// <summary>
    /// Writes enqueued LogItem entries to file in a background thread.
    /// </summary>
    public static class DebugLogWriter
    {
        static bool enabled = false;
        static BlockingCollection<DebugLogItem?> items = new BlockingCollection<DebugLogItem?>();
        static TextWriter output;
        static Stream LastFileOutputStream;
        static Task processTask;
        static object lockobject = new object();

        /// <summary>
        /// Initializes the DebugLogWriter with new parameters
        /// </summary>
        /// <param name="logPath">path to log file on disk</param>
        /// <param name="append">indicates new logs should be written to an already existing log file</param>
        /// <param name="maxLogSize">max. log file size in bytes</param>
        public static void Init(string logPath, bool append, long maxLogSize)
        {
            if ( enabled ) return;

            // close formerly opened stream
            if ( LastFileOutputStream != null )
                LastFileOutputStream.Close();
            LastFileOutputStream = null;

            // open new stream
            string logDirectory = Path.GetDirectoryName(logPath);
            if ( !Directory.Exists(logDirectory) )
            {
                Directory.CreateDirectory(logDirectory);
            }

            Exception truncateFailed = null;
            // ensure that the logfile doesn't exceed a maximum size
            if ( append & File.Exists(logPath) )
            {
                if ( maxLogSize > 0 && new FileInfo(logPath).Length > maxLogSize )
                {
                    try
                    {
                        TruncateLogfile(logPath, maxLogSize / 2);
                    }
                    catch ( Exception ex )
                    {
                        truncateFailed = ex;
                    }
                }
            }

            LastFileOutputStream = File.Open(logPath, append ? FileMode.Append : FileMode.Create, FileAccess.Write, FileShare.Read);
            output = new StreamWriter(LastFileOutputStream);
            items = new BlockingCollection<DebugLogItem?>();

            // start tast that writes new items to file
            processTask = Task.Factory.StartNew(() => Process(), TaskCreationOptions.LongRunning);

            // mark as active and working
            enabled = true;
        }

        /// <summary>
        /// Removes the beginning from a text file until at most keepMax bytes are left
        /// </summary>
        private static void TruncateLogfile(string fName, long keepMax)
        {
            string bakName = Path.ChangeExtension(fName, "__bak__");
            using ( var rdr = File.OpenText(fName) )
            {
                rdr.BaseStream.Seek(-keepMax, SeekOrigin.End);
                rdr.ReadLine();
                using ( var wr = File.CreateText(bakName) )
                {
                    string tmp = rdr.ReadLine();
                    while ( tmp != null )
                    {
                        wr.WriteLine(tmp);
                        tmp = rdr.ReadLine();
                    }
                }
            }
            File.Delete(fName);
            File.Move(bakName, fName);
        }

        /// <summary>
        /// Enqueues new item to be written to file.
        /// </summary>
        public static void Enqueue(DebugLogItem? logItem)
        {
            if ( enabled && logItem != null )
            {
                items.Add(logItem);
            }
        }

        public static void Enqueue(string format, params object[] entries)
        {
            if ( enabled )
            {
                items.Add(new DebugLogItem(format, entries));
            }
        }

        /// <summary>
        /// forces a LogItem to be written to the logfile instantly, bypassing the log-thread
        /// </summary>
        /// <param name="item">LogItem to be written</param>
        public static void ForceWrite(DebugLogItem? item)
        {
            if ( item == null )
                return;

            lock ( lockobject )
            {
                if ( output != null )
                {
                    output.WriteLine(item.ToString());
                    output.Flush();
                }
            }
        }

        /// <summary>
        /// Writes items from queue to file
        /// </summary>
        private static void Process()
        {
            while ( true )
            {
                DebugLogItem? item;
                if ( items.TryTake(out item, 10) )
                {
                    try
                    {
                        if ( item == null )     // signals the end of this Task
                            return;
                        lock ( lockobject )
                        {
                            if ( output != null )
                            {
                                output.WriteLine(item.ToString());
                                output.Flush();
                            }
                        }
                    }
#if DEBUG
                    catch ( Exception ex )
                    {
                        Console.WriteLine("ERROR: Could not write item to log file: {0}", ex.Message); ;
                    }
#else
                    catch {}
#endif
                }
            }
        }

        /// <summary>
        /// Shuts down background worker threads and disposes resources.
        /// </summary>
        public static void Dispose()
        {
            enabled = false;               // no more Enqueues allowed 

            try
            {
                items.Add(null);           // signal task end!
                if ( processTask != null )
                {
                    processTask.Wait(1000);    // wait for task to complete
                    processTask.Dispose();     // kill task
                }
            }
#if DEBUG
            catch ( Exception ex )
            {
                Console.WriteLine("ERROR: Could not shutdown TextWriter task: {0}", ex.Message); ;
            }
#else
            catch {}
#endif
            if ( LastFileOutputStream != null )
                LastFileOutputStream.Close();       // close file stream
        }
    }

    #endregion


    #region namespace Win32

    namespace Win32
    {
        // see:
        // https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-sendinput
        // https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-initializetouchinjection
        // https://github.com/michaelnoonan/inputsimulator
        // https://github.com/michaelosthege/TouchInjector
        // https://social.technet.microsoft.com/wiki/contents/articles/6460.simulating-touch-input-in-windows-8-using-touch-injection-api.aspx?PageIndex=2


        public partial class NativeMethods
        {
            /// <summary>
            /// The SendInput function synthesizes keystrokes, mouse motions, and button clicks.
            /// </summary>
            /// <param name="numberOfInputs">Number of structures in the Inputs array.</param>
            /// <param name="inputs">Pointer to an array of INPUT structures. Each structure represents an event to be inserted into the keyboard or mouse input stream.</param>
            /// <param name="sizeOfInputStructure">Specifies the size, in bytes, of an INPUT structure. If cbSize is not the size of an INPUT structure, the function fails.</param>
            /// <returns>The function returns the number of events that it successfully inserted into the keyboard or mouse input stream. If the function returns zero, the input was already blocked by another thread. To get extended error information, call GetLastError.Microsoft Windows Vista. This function fails when it is blocked by User Interface Privilege Isolation (UIPI). Note that neither GetLastError nor the return value will indicate the failure was caused by UIPI blocking.</returns>
            /// <remarks>
            /// Microsoft Windows Vista. This function is subject to UIPI. Applications are permitted to inject input only into applications that are at an equal or lesser integrity level.
            /// The SendInput function inserts the events in the INPUT structures serially into the keyboard or mouse input stream. These events are not interspersed with other keyboard or mouse input events inserted either by the user (with the keyboard or mouse) or by calls to keybd_event, mouse_event, or other calls to SendInput.
            /// This function does not reset the keyboard's current state. Any keys that are already pressed when the function is called might interfere with the events that this function generates. To avoid this problem, check the keyboard's state with the GetAsyncKeyState function and correct as necessary.
            /// </remarks>
            [DllImport("user32.dll", SetLastError = true)]
            internal static extern UInt32 SendInput(UInt32 numberOfInputs, INPUT[] inputs, int sizeOfInputStructure);

            /// <summary>
            /// Init. call for Touch Injection.
            /// See: https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-initializetouchinjection
            /// </summary>
            /// <param name="maxCount">The maximum number of touch contacts.</param>
            /// <param name="mode">see TOUCH_FEEDBACK_INDIRECT</param>
            /// <returns></returns>
            [DllImport("user32.dll", SetLastError = true)]
            //[return: MarshalAs(UnmanagedType.Bool)]
            internal static extern bool InitializeTouchInjection(UInt32 maxCount, int mode);

            /// <summary>
            /// Inject Touch Information.
            /// See: https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-injecttouchinput
            /// </summary>
            /// <param name="count"></param>
            /// <param name="touchInfo">array of POINTERTOUCHINFO structures</param>
            /// <returns></returns>
            [DllImport("user32.dll", SetLastError = true)]
            //[return: MarshalAs(UnmanagedType.Bool)]
            internal static extern bool InjectTouchInput(UInt32 count, [MarshalAs(UnmanagedType.LPArray), In] POINTERTOUCHINFO[] touchInfo);

            [DllImport("kernel32.dll")]
            internal static extern uint GetLastError();
        }

        [StructLayout(LayoutKind.Sequential)]
        public struct POINT
        {
            public int x;
            public int y;
        }

        [StructLayout(LayoutKind.Explicit)]
        public struct RECT
        {
            [FieldOffset(0)]
            public int left;
            [FieldOffset(4)]
            public int top;
            [FieldOffset(8)]
            public int right;
            [FieldOffset(12)]
            public int bottom;
        }

        // see: https://docs.microsoft.com/en-us/windows/win32/api/winuser/ns-winuser-input
        public struct INPUT
        {
            public UInt32 Type;     // see SendInputEventType enum
            public MOUSEKEYBDHARDWAREINPUT Data;
        }

        // see: https://docs.microsoft.com/en-us/windows/win32/api/winuser/ns-winuser-input
        [StructLayout(LayoutKind.Explicit)]
        public struct MOUSEKEYBDHARDWAREINPUT
        {
            /// <summary>
            /// The <see cref="MOUSEINPUT"/> definition.
            /// </summary>
            [FieldOffset(0)]
            public MOUSEINPUT Mouse;

            /// <summary>
            /// The <see cref="KEYBDINPUT"/> definition.
            /// </summary>
            [FieldOffset(0)]
            public KEYBDINPUT Keyboard;

            /// <summary>
            /// The <see cref="HARDWAREINPUT"/> definition.
            /// </summary>
            [FieldOffset(0)]
            public HARDWAREINPUT Hardware;
        }

        // see: https://docs.microsoft.com/en-us/windows/win32/api/winuser/ns-winuser-pointer_touch_info
        //[StructLayout(LayoutKind.Sequential)]
        public struct POINTERTOUCHINFO
        {
            // Contains basic pointer information common to all pointer types.
            public POINTERINFO pointerInfo;

            // Lists the touch flags.
            public TOUCHFLAGS touchFlags;

            public uint touchMasks;

            // Pointer contact area in pixel screen coordinates. 
            // By default, if the device does not report a contact area, 
            // this field defaults to a 0-by-0 rectangle centered around the pointer location.
            public RECT rcContact;

            // The raw screen coordinates of the contact area, in pixels. For adjusted screen coordinates, see 'rcContact'
            public RECT rcContactRaw;

            // A pointer orientation, with a value between 0 and 359, where 0 indicates a touch pointer 
            // aligned with the x-axis and pointing from left to right; increasing values indicate degrees
            // of rotation in the clockwise direction.
            // This field defaults to 0 if the device does not report orientation.
            public uint orientation;

            // Pointer pressure normalized in a range of 0 to 256.
            public uint pressure;
        }

        // see: https://docs.microsoft.com/en-us/windows/win32/api/winuser/ns-winuser-pointer_info
        //[StructLayout(LayoutKind.Sequential)]
        public struct POINTERINFO
        {
            public POINTERINPUTTYPE pointerType;
            public uint pointerId;
            public uint frameId;
            public POINTERFLAGS pointerFlags;
            public IntPtr sourceDevice;
            public IntPtr hwndTarget;
            public POINT ptPixelLocation;
            public POINT ptHimetricLocation;
            public POINT ptPixelLocationPredicted;
            public POINT ptHimetricLocationPredicted;
            public uint dwTime;
            public uint historyCount;
            public uint inputData;
            public uint dwKeyStates;
            public ulong PerformanceCount;
            public POINTERBUTTONCHANGETYPE ButtonChangeType;
        }

        // see: https://docs.microsoft.com/en-us/previous-versions/windows/desktop/inputmsg/touch-flags-constants
        public enum TOUCHFLAGS
        {
            TOUCH_FLAGS_NONE = 0x00000000/*Indicates that no flags are set.*/
        }

        // see: https://docs.microsoft.com/en-us/previous-versions/windows/desktop/inputmsg/touch-mask-constants
        [Flags]
        public enum TOUCHMASK
        {
            TOUCH_MASK_NONE = 0x00000000,
            TOUCH_MASK_CONTACTAREA = 0x00000001,
            TOUCH_MASK_ORIENTATION = 0x00000002,
            TOUCH_MASK_PRESSURE = 0x00000004
        }

        // see: https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-initializetouchinjection
        public enum TOUCHFEEDBACKTYPE : uint
        {
            TOUCH_FEEDBACK_DEFAULT = 0x001,
            TOUCH_FEEDBACK_INDIRECT = 0x002,
            TOUCH_FEEDBACK_NONE = 0x003,
        }

        // see: https://docs.microsoft.com/en-us/windows/win32/api/winuser/ne-winuser-tagpointer_input_type
        public enum POINTERINPUTTYPE
        {
            PT_POINTER = 0x00000001,
            PT_TOUCH = 0x00000002,
            PT_PEN = 0x00000003,
            PT_MOUSE = 0x00000004
        };

        // see: https://docs.microsoft.com/en-us/previous-versions/windows/desktop/inputmsg/pointer-flags-contants
        public enum POINTERFLAGS
        {
            POINTER_FLAG_NONE = 0x00000000,
            POINTER_FLAG_NEW = 0x00000001,
            POINTER_FLAG_INRANGE = 0x00000002,
            POINTER_FLAG_INCONTACT = 0x00000004,
            POINTER_FLAG_FIRSTBUTTON = 0x00000010,
            POINTER_FLAG_SECONDBUTTON = 0x00000020,
            POINTER_FLAG_THIRDBUTTON = 0x00000040,
            POINTER_FLAG_OTHERBUTTON = 0x00000080,
            POINTER_FLAG_PRIMARY = 0x00000100,
            POINTER_FLAG_CONFIDENCE = 0x00000200,
            POINTER_FLAG_CANCELLED = 0x00000400,
            POINTER_FLAG_DOWN = 0x00010000,
            POINTER_FLAG_UPDATE = 0x00020000,
            POINTER_FLAG_UP = 0x00040000,
            POINTER_FLAG_WHEEL = 0x00080000,
            POINTER_FLAG_HWHEEL = 0x00100000
        }

        // see: https://docs.microsoft.com/en-us/windows/win32/api/winuser/ne-winuser-pointer_button_change_type
        public enum POINTERBUTTONCHANGETYPE
        {
            POINTER_CHANGE_NONE,
            POINTER_CHANGE_FIRSTBUTTON_DOWN,
            POINTER_CHANGE_FIRSTBUTTON_UP,
            POINTER_CHANGE_SECONDBUTTON_DOWN,
            POINTER_CHANGE_SECONDBUTTON_UP,
            POINTER_CHANGE_THIRDBUTTON_DOWN,
            POINTER_CHANGE_THIRDBUTTON_UP,
            POINTER_CHANGE_FOURTHBUTTON_DOWN,
            POINTER_CHANGE_FOURTHBUTTON_UP,
            POINTER_CHANGE_FIFTHBUTTON_DOWN,
            POINTER_CHANGE_FIFTHBUTTON_UP
        }

        /// <summary>
        /// Contains information about a simulated keyboard event.
        /// See: https://docs.microsoft.com/en-us/windows/win32/api/winuser/ns-winuser-keybdinput
        /// </summary>
        public struct KEYBDINPUT
        {
            /// <summary>
            /// Specifies a virtual-key code. The code must be a value in the range 1 to 254. The Winuser.h header file provides macro definitions (VK_*) for each value. If the dwFlags member specifies KEYEVENTF_UNICODE, wVk must be 0. 
            /// </summary>
            public UInt16 KeyCode;

            /// <summary>
            /// Specifies a hardware scan code for the key. If dwFlags specifies KEYEVENTF_UNICODE, wScan specifies a Unicode character which is to be sent to the foreground application. 
            /// </summary>
            public UInt16 Scan;

            /// <summary>
            /// Specifies various aspects of a keystroke. This member can be certain combinations of the following values.
            /// KEYEVENTF_EXTENDEDKEY - If specified, the scan code was preceded by a prefix byte that has the value 0xE0 (224).
            /// KEYEVENTF_KEYUP - If specified, the key is being released. If not specified, the key is being pressed.
            /// KEYEVENTF_SCANCODE - If specified, wScan identifies the key and wVk is ignored. 
            /// KEYEVENTF_UNICODE - Windows 2000/XP: If specified, the system synthesizes a VK_PACKET keystroke. The wVk parameter must be zero. This flag can only be combined with the KEYEVENTF_KEYUP flag. For more information, see the Remarks section. 
            /// </summary>
            public UInt32 Flags;

            /// <summary>
            /// Time stamp for the event, in milliseconds. If this parameter is zero, the system will provide its own time stamp. 
            /// </summary>
            public UInt32 Time;

            /// <summary>
            /// Specifies an additional value associated with the keystroke. Use the GetMessageExtraInfo function to obtain this information. 
            /// </summary>
            public IntPtr ExtraInfo;
        }

        /// <summary>
        /// Contains information about a simulated mouse event.
        /// Declared in Winuser.h, include Windows.h
        /// See: https://docs.microsoft.com/en-us/windows/win32/api/winuser/ns-winuser-mouseinput
        /// </summary>
        public struct MOUSEINPUT
        {
            /// <summary>
            /// Specifies the absolute position of the mouse, or the amount of motion since the last mouse event was generated, depending on the value of the dwFlags member. Absolute data is specified as the x coordinate of the mouse; relative data is specified as the number of pixels moved. 
            /// </summary>
            public int X;

            /// <summary>
            /// Specifies the absolute position of the mouse, or the amount of motion since the last mouse event was generated, depending on the value of the dwFlags member. Absolute data is specified as the y coordinate of the mouse; relative data is specified as the number of pixels moved. 
            /// </summary>
            public int Y;

            /// <summary>
            /// If dwFlags contains MOUSEEVENTF_WHEEL, then mouseData specifies the amount of wheel movement. A positive value indicates that the wheel was rotated forward, away from the user; a negative value indicates that the wheel was rotated backward, toward the user. One wheel click is defined as WHEEL_DELTA, which is 120. 
            /// Windows Vista: If dwFlags contains MOUSEEVENTF_HWHEEL, then dwData specifies the amount of wheel movement. A positive value indicates that the wheel was rotated to the right; a negative value indicates that the wheel was rotated to the left. One wheel click is defined as WHEEL_DELTA, which is 120.
            /// Windows 2000/XP: IfdwFlags does not contain MOUSEEVENTF_WHEEL, MOUSEEVENTF_XDOWN, or MOUSEEVENTF_XUP, then mouseData should be zero. 
            /// If dwFlags contains MOUSEEVENTF_XDOWN or MOUSEEVENTF_XUP, then mouseData specifies which X buttons were pressed or released. This value may be any combination of the following flags. 
            /// </summary>
            public UInt32 MouseData;

            /// <summary>
            /// A set of bit flags that specify various aspects of mouse motion and button clicks. The bits in this member can be any reasonable combination of the following values. 
            /// The bit flags that specify mouse button status are set to indicate changes in status, not ongoing conditions. For example, if the left mouse button is pressed and held down, MOUSEEVENTF_LEFTDOWN is set when the left button is first pressed, but not for subsequent motions. Similarly, MOUSEEVENTF_LEFTUP is set only when the button is first released. 
            /// You cannot specify both the MOUSEEVENTF_WHEEL flag and either MOUSEEVENTF_XDOWN or MOUSEEVENTF_XUP flags simultaneously in the dwFlags parameter, because they both require use of the mouseData field. 
            /// </summary>
            public UInt32 Flags;

            /// <summary>
            /// Time stamp for the event, in milliseconds. If this parameter is 0, the system will provide its own time stamp. 
            /// </summary>
            public UInt32 Time;

            /// <summary>
            /// Specifies an additional value associated with the mouse event. An application calls GetMessageExtraInfo to obtain this extra information. 
            /// </summary>
            public IntPtr ExtraInfo;
        }

        /// <summary>
        /// Contains information about a simulated message generated by an input device other than a keyboard or mouse.
        /// Declared in Winuser.h, include Windows.h    
        /// see: https://docs.microsoft.com/en-us/windows/win32/api/winuser/ns-winuser-hardwareinput
        /// </summary>
        public struct HARDWAREINPUT
        {
            /// <summary>
            /// Value specifying the message generated by the input hardware. 
            /// </summary>
            public UInt32 Msg;

            /// <summary>
            /// Specifies the low-order word of the lParam parameter for uMsg. 
            /// </summary>
            public UInt16 ParamL;

            /// <summary>
            /// Specifies the high-order word of the lParam parameter for uMsg. 
            /// </summary>
            public UInt16 ParamH;
        }

        // see: https://docs.microsoft.com/en-us/windows/win32/api/winuser/ns-winuser-keybdinput
        [Flags]
        public enum KEYBOARDFLAGS : uint // UInt32
        {
            /// <summary>
            /// KEYEVENTF_EXTENDEDKEY = 0x0001 (If specified, the scan code was preceded by a prefix byte that has the value 0xE0 (224).)
            /// </summary>
            ExtendedKey = 0x0001,

            /// <summary>
            /// KEYEVENTF_KEYUP = 0x0002 (If specified, the key is being released. If not specified, the key is being pressed.)
            /// </summary>
            KeyUp = 0x0002,

            /// <summary>
            /// KEYEVENTF_UNICODE = 0x0004 (If specified, wScan identifies the key and wVk is ignored.)
            /// </summary>
            Unicode = 0x0004,

            /// <summary>
            /// KEYEVENTF_SCANCODE = 0x0008 (Windows 2000/XP: If specified, the system synthesizes a VK_PACKET keystroke. The wVk parameter must be zero. This flag can only be combined with the KEYEVENTF_KEYUP flag. For more information, see the Remarks section.)
            /// </summary>
            ScanCode = 0x0008,
        }

        // see: https://docs.microsoft.com/en-us/windows/win32/api/winuser/ns-winuser-mouseinput
        [Flags]
        public enum MOUSEFLAGS : uint
        {
            MOUSEEVENT_MOVE = 0x0001,
            MOUSEEVENT_LEFTDOWN = 0x0002,
            MOUSEEVENT_LEFTUP = 0x0004,
            MOUSEEVENT_RIGHTDOWN = 0x0008,
            MOUSEEVENT_RIGHTUP = 0x0010,
            MOUSEEVENT_MIDDLEDOWN = 0x0020,
            MOUSEEVENT_MIDDLEUP = 0x0040,
            MOUSEEVENT_XDOWN = 0x0080,
            MOUSEEVENT_XUP = 0x0100,
            MOUSEEVENT_WHEEL = 0x0800,
            MOUSEEVENT_VIRTUALDESK = 0x4000,
            MOUSEEVENT_ABSOLUTE = 0x8000
        }

        // see: https://docs.microsoft.com/en-us/windows/win32/api/winuser/ns-winuser-input
        [Flags]
        public enum SENDINPUTEVENTTYPE : uint
        {
            InputMouse,
            InputKeyboard,
            InputHardware
        }
    }

    #endregion
}
